'''
mcreep.model
------------
Definition of Model for fitting the creep data + functions to run it.

The module defines Model class, whose object/instance contains:

* Properties for fitting model to data:
    1. property func: definition of fitting function
    2. property table_of_results: for saving the results of fitting
    3. and a few other properties, which can modify the fitting prodedure.
* Functions/methods for fitting model to data - the most important are:
    1. Model.run, which runs the whole calculation
    2. Model.describe, which briefly describes the model (for outputs, reports)
    3. Model.final_report, which creates final reports from ALL fitting results
* Three objects describing the input/output data:
    1.  mcreep.const.Experiment = description of experiment
    2.  mcreep.const.DataParameters = description of input data
    3.  mcreep.const.PlotParameters = description of the output plot
'''

import sys
import math
import pandas as pd
from pathlib import Path
import mcreep.io, mcreep.fit

class Model:
    '''
    Model for fitting creep data, containing namely:
        
    * fitting function (a function selected from mcreep.func)
    * methods to run the fitting (employing also mcreep.io, mcreep.fit)
    * table/dataframe with results of fitting; updated during calculations
    
    Parameters
    ----------
    EPAR : mcreep.const.Experiment object
        Description of experimental parameters (such as experiment type...).
    
    DPAR : mcreep.const.DataParameters object
        Description of data (such as which columns to read, rows to skip...)
    
    PPAR : mcreep.const.PlotParameters object
        Definition of plot parameters (xlabel, ylabel, global properties...).
    
    name : str
        Name of the fitting function in human-readable form.
    
    func : function object
        One of the fitting functions defined in module mcreep.func,
        such as `mcreep.func.power_law`, `mcreep.func.evp_s_d_1kv`, ...
    
    rtimes : list of floats; optional, default is None
        List of retardation times.
        If defined, the retardation times of EVP fitting functions
        are fixed at given values.
        
    iguess : list of floats; default is None
        Initial guess = list of starting values for fitting.
        The number of values must correspond to variables in given model.
        If defined, iguess parameter is passed to scipy.optimize.curve_fit
    
    print_covariances : bool; default is False
        If true, print covariance matrix.
        The diagonal elements of covariance matrix are `sigma**2(par1,par1)`
        => if they are close to zero, sigma of par1 is very low.
        The off-diagonal elements of covariance matrix = `sigma**2(par1,par2)`
        => if they are zero, par1 and par2 are independent on each other.
        Additional theory concerning the covariation matrix
        can be found elsewhere: https://stats.stackexchange.com/q/151018
        
    output_dir : str or path-like object; optional, default is '.'
        Directory for output graphs = PNG-files.
        If not given, we use current directory,
        i.e. the directory of the script that runs the program.
        We note that final report = TXT-file
        is always written in current directory by default,
        unless we call Model.final_report method with output_file argument.

    Returns
    -------
    Model object :  
        * The Model.__init__ function
          returns the object with all properties and methods
        * The Model.run method
          fits the selected model function to experimental creep data
          and saves the results to {Model.table_of_results} property.
    
    Additional parameters
    ---------------------
    * table_of_results : pandas.DataFrame
        - Table of fitting results, which contains
          datafile name + values of fitted variables.
    * table_of_results_evp : pandas.DataFrame
        - Table of recalculated fitting results for EVP models, containing:
            * final compliances: `C0,Cv,C1,C2...`
            * final retardation times: `tau1,tau2...`
        - These values can be used to predict E(t):
            * `E(t) = f(C0,Cv,C1...,tau1...)`
    '''
    
    def __init__(self, EPAR, DPAR, PPAR, name, func, 
                 rtimes=None, iguess=None, print_covariances=False,
                 output_dir='.'):
        # Docstring for __init__ are given above in class description.
        # Reason: In this way, the parameters are visible in Spyder/Ctrl+I.

        # (1) Read basic properties.
        # (read during initialization
        self.EPAR = EPAR
        self.DPAR = DPAR
        self.PPAR = PPAR
        self.name = name
        self.func = func
        self.rtimes = rtimes
        self.iguess = iguess
        self.print_covariances = print_covariances
        self.output_dir = Path(output_dir)
        # (2) Modify/initialize additional properties
        # `(a) fname = name of the function; needed in future modifications
        self.fname = func.__name__
        # (b) results = pandas.Dataframes that keeps the results of fitting
        self._initialize_tables_of_results()
        # (3) Fix constants in EVP functions 
        # (a) fix basic constant (multiplicative constant in EVP models)
        if self.fname.startswith('evp'):
            self.func = self._evp_fix_basic_constants(EPAR)
        # (b) fix retardation times if requested => if rtimes != None
        if self.fname.startswith('evp') and rtimes != None:
            self.func = self._evp_fix_retardation_times(rtimes)
    
    def _initialize_tables_of_results(self):
        '''
        Initilize two tables,
        into which we are going to save fitting results.

        Returns
        -------
        None
            The result is correct initialization of two tables:
            table_of_results = for all models,
            table_of_results_evp = extra table for EVP models.

        '''
        # (1) Prepare columns = names of parameters for different models
        if self.fname == 'power_law':
            table1 = ['C','n']
            table2 = None
        elif self.fname == 'nutting_law':
            table1 = ['e0','C','n']
            table2 = None
        elif self.fname == 'evp_s_d_1kv':
            table1 = ['const','B0','Cv','D1','tau1']
            table2 = ['C0','Cv','C1','tau1']
        elif self.fname == 'evp_s_d_2kv':
            table1 = ['const','B0','Cv','D1','D2','tau1','tau2']
            table2 = ['C0','Cv','C1','C2','tau1','tau2']
        elif self.fname == 'evp_s_d_3kv':
            table1 = ['const','B0','Cv','D1','D2','D3','tau1','tau2','tau3']
            table2 = ['C0','Cv','C1','C2','C3','tau1','tau2','tau3']
        else:
            sys.exit('Unknown model!')
        # (2) Two additional columns for statistics
        # R2fit = coefficient of determination for fitted = [t_fstart; t_fend]
        # R2all = coefficient of determination for all data = [t_start; t_end]
        table1.extend(['R2fit','R2all']) 
        # (3) Create empty tables with prepared columns
        self.table_of_results = pd.DataFrame(columns=table1)
        self.table_of_results_evp = pd.DataFrame(columns=table2)            
            
    def _evp_fix_basic_constants(self, EPAR):
        '''
        Fix basic constants in EVP functions.

        Parameters
        ----------
        self : Model object
            Object with model fitting function.
        
        EPAR : Experiment object.
            Object with experimental data, including constants (F,K).
        
        Returns
        -------
        func_with_fixed_constants : function object
            The original function, but with fixed/constant parameters (F,K).
        
        Notes
        -----
        * SciPy algorithm for curve fitting `scipy.optimize.curve_fit`
          does not take into account keyword/fixed/constant arguments,
          i.e. it always fits all arguments regardless if they have value.
        * Therefore, the function has to be modified
          so that the keyword arguments are replaced by the values
          and removed from the fitting function definition.
        * https://stackoverflow.com/q/12208634
        
        '''
        # (1) Initialize
        # The following assignment is necessary.
        # If we use self.func directly, it causes mysterious errors.
        func = self.func
        # (2) Change functions = fix constants = eliminate constant arguments.
        # (here we just change the initial multiplicative constant
        # (the following assignement is just for convenience
        const = self.EPAR.const
        # Now we eliminate the const argument from the function definition...
        if self.fname == 'evp_s_d_1kv':
            func_with_fixed_constants = \
                lambda t,B0,Cv,D1,tau1 : \
                    func(t,const,B0,Cv,D1,tau1)
        elif self.fname == 'evp_s_d_2kv':
            func_with_fixed_constants = \
                lambda t,B0,Cv,D1,D2,tau1,tau2 : \
                    func(t,const,B0,Cv,D1,D2,tau1,tau2)
        elif self.fname == 'evp_s_d_3kv':
            func_with_fixed_constants = \
                lambda t,B0,Cv,D1,D2,D3,tau1,tau2,tau3 : \
                    func(t,const,B0,Cv,D1,D2,D3,tau1,tau2,tau3)
        else:
            # Non-EVP function => nothing to fix...
            func_with_fixed_constants = self.func
        # (3) Keep the name of original function
        # (this is not necessary here, fname is saved in self.fname
        # (however, it is "more correct" and may be necessary in non-OO code
        func_with_fixed_constants.__name__ = self.fname
        # (4) Return final function with fixed constants
        return(func_with_fixed_constants)
   
    def _evp_fix_retardation_times(self, rtimes):
        '''
        Fix retardation times in EVP functions
        if argument rtimes is not empty.

        Parameters
        ----------
        rtimes : list of floats
            Fixed values of retardation times of EVP models = tau1,tau2...
            The number of values must correspond to given model, of course.

        Returns
        -------
        None
        
        Notes
        -----
        * SciPy algorithm for curve fitting `scipy.optimize.curve_fit`
          does not take into account keyword/fixed/constant arguments,
          i.e. it always fits all arguments regardless if they have value.
        * Therefore, the function has to be modified
          so that the retardation times are replaced by constants
          and removed from the fitting function definition
          (if they are fixed by means of argument rtimes).
        * https://stackoverflow.com/q/12208634
        
        '''
        # (1) Initialize
        # The following assignment is necessary.
        # If we use self.func directly, it causes mysterious errors.
        func = self.func
        # (2) Change functions = fix constants = eliminate constant arguments.
        if self.fname == 'evp_s_d_1kv':
            func_with_fixed_rtimes = \
                lambda t,B0,Cv,D1 : \
                    func(t,B0,Cv,D1,rtimes[0])
        elif self.fname == 'evp_s_d_2kv':
            func_with_fixed_rtimes = \
                lambda t,B0,Cv,D1,D2 : \
                    func(t,B0,Cv,D1,D2,rtimes[0],rtimes[1])
        elif self.fname == 'evp_s_d_3kv':
            func_with_fixed_rtimes = \
                lambda t,B0,Cv,D1,D2,D3 : \
                    func(t,B0,Cv,D1,D2,D3,rtimes[0],rtimes[1],rtimes[2])
        else:
            # Non-EVP function => nothing to fix...
            func_with_fixed_rtimes = self.func
        # (3) Keep the name of original function
        # (this is not necessary here, fname is saved in self.fname
        # (however, it is "more correct" and may be necessary in non-OO code
        func_with_fixed_rtimes.__name__ = self.fname
        # (4) Return final function with fixed constants
        return(func_with_fixed_rtimes)
    
    def describe(self, fh=sys.stdout):
        '''
        Print brief description of the fitting model and its outputs.

        Parameters
        ----------
        fh : filehandle; optional, default is sys.stdout
            By default, the model description is printed on stdout.
            If {fh} is given, the output is redirected to {fh} filehandle.
            Reason: the description can be printed both to stdout
            and to text file.

        Returns
        -------
        None
            The output is the text printed to stdout or text file.
        '''
        # (0) If {fh} is given redirect standard output to filehandle.
        # (this is a dirty trick, but it simplifies/shortens code a bit
        if fh != None:
            original_standard_output = sys.stdout
            sys.stdout = fh
        # (1) Print description
        # (either to sys.stdout or to sys.stdout = fh
        # ...brief info about experiment/measurement type.
        if self.EPAR.etype == 'Tensile':
            print('Tensile creep experiment.')
        elif self.EPAR.etype == 'Vickers':
            print('Indentation creep with Vickers tip.')
        elif self.EPAR.etype == 'Berkovich':
            print('Indentation creep with Berkovich tip.')
        elif self.EPAR.etype == 'Spherical':
            print('Indentation creep with Spherical tip.')
        # ...information about the model and results.
        if self.fname == 'power_law':
            print('Model function: Power Law => deformation(t) = C * t**n')
            print('Units are relative; n = creep constant ~ creep rate.')
        elif self.fname == 'nutting_law':
            print("Model function: Nutting's Law => def(t) = e0 + C * t**n")
            print("Units are relative; n = creep constant ~ creep rate.")
        elif self.fname.startswith('evp'):
            print('Model function: EVP with S,D and KV components.')
            print('Compliances B,C,D in [GPa], retardation times tau in [s].')
        # ...final end-of-line
        print()
        # (2) If {fh} was used, restore original standard output...
        if fh != None:
            sys.stdout = original_standard_output
    
    def run(self, datafile, t_start, t_hold, t_fstart=None, t_fend=None):
        '''
        Run the model = fit model functin to experimental data.
        
        * This includes reading data + fitting + plotting + saving results.
        * The method employs other functions defined mcreep.io + mcreep.fit.

        Parameters
        ----------
        datafile : str or path-like object
            Full name of the datafile containing creep data.
        
        t_start : float
            The creep data are read for the interval
            [t_start; t_start+t_hold];
            t_start is the initial time for the data reading.
                For indentation experiments, t_start ~ when Fmax is reached.
                For tensile experiments, t_start ~ the first detected time.
                Alternatively, t_start can be a bit higher
                in order to ignore the initial period.
        
        t_hold : float
            The creep data are read for the interval
            [t_start; t_start+t_hold];
            t_hold is used to calculate the final time for the data reading.
                For both indentation and tensile experiments,
                it is the time for which the Fmax is held.
                It is reasonable to insert a slightly lower time
                (to be safe + to consider possible increase in t_start).
        
        t_fstart : float; optional
            The model function is fitted to creep data in the interval
            [t_fstart; t_fend].
            If t_fstart is not given, it is set equal to t_start;
            i.e. the function is fitted to all data that were read.
            
        t_fend : float; optional
            The model function is fitted to creep data in the interval
            [t_fstart; t_fend].
            If t_fstart is not given, it is set equal to (t_start + t_hold);
            i.e. the function is fitted to all data that were read.

        Returns
        -------
        None
            Nevertheless, this is the key method of {Model} object which...
            
            * reads creep data
            * fits model function to creep data
            * saves results to self.table_of_results
              and self.table_of_results_evp
            * Note: the printing and saving data to file
              is performed after running the model (self.run)
              by means of another method (self.final_report).
        '''
        # (0) Set t_fstart,t_fend to defaults,
        # if they were not given as arguments.
        if t_fstart == None: t_fstart = t_start
        if t_fend == None: t_fend = t_start+t_hold
        # (1) Read datafile = experimental data.
        data = self.read_datafile(datafile, t_start, t_hold)
        # (2) Fit datafile/experimental data with model function.
        par,cov = self.fit_function_to_data(data, t_fstart, t_fend)
        # (3) Print the result of fitting.
        # (Note: we convert datafile to pure/string name without path
        # (Reason: datafile might have been gi ven as pathlib object...
        # (...and the pathlib object cannot be printed easily as string
        datafile = Path(datafile).name
        self.print_fitting_result(datafile, par, cov)
        # (3) Plot the result of fitting.
        self.plot_fitting_result(datafile, data, par)
        # (4) Calculate statistics
        R2fit, R2all = self.calculate_statistics(par, data, t_fstart, t_fend)
        # (5) Recalculate and save the result of fitting to MODEL object
        # (the results are saved in both original and recalculated form
        # (original results = fitting/regression parameters
        # (recalculated results - for EVP models: B,D => C = compliances
        # (EVP recalculations consider also difference tensile vs. indentation
        self.recalc_and_save_fitting_results( 
            datafile, par, R2fit, R2all, t_start)

    def read_datafile(self, datafile, t_start, t_hold):
        '''
        Read datafile with creep data
        (just a wrapper for function mcreep.io.read_datafile).
        '''
        data = mcreep.io.read_datafile(self, datafile, t_start, t_hold)
        return(data)
    
    def fit_function_to_data(self, data, t_fstart, t_fend):
        '''
        Fit model function to creep data
        (just a wrapper for function mcreep.fit.fit).
        '''
        par,cov = mcreep.fit.fit(self, data, t_fstart, t_fend)
        return(par, cov)
    
    def print_fitting_result(self, datafile, par, cov):
        '''
        Print the results of creep data fitting with given model.
        (just a wrapper for function mcreep.io.print_fitting_result).
        '''
        mcreep.io.print_fitting_result(self, datafile, par, cov)
    
    def plot_fitting_result(self, datafile, data, par):
        '''
        Save the results of creep data fitting with given model.
        (just a wrapper for function mcreep.io.plot_fitting_result).
        '''
        mcreep.io.plot_fitting_result(self, datafile, data, par)
    
    def calculate_statistics(self, par, data, t_fstart, t_fend):
        '''
        Calculate statistics for model function fitted to creep data.
        It calls mcreep.fit.coefficient_of_determination twice:
        
        * once for all data (= data in interval [t_start, t_start+t_hold])
        * once for fitted data (= data in interval [t_fstart, t_fend])
        
        Parameters
        ----------
        par : list of floats
            Result of fitting procedure;
            usually the result of procedure mcreep.fit.fit.
        
        data : 2D-numpy array
            Crep data read from input datafile;
            usually the result of procedure mcreep.io.read_datafile.
            
        t_fstart, t_fend : float, float
            Times determining the fitting interval = [t_fstart; t_fend].
        
        Returns
        -------
        R2fit, R2all : float, float
            Coefficients of determination for data in fitting interval
            [t_fstart, t_fend] and for the whole dataset in interval
            [t_start, t_start+t_hold].
        
        '''
        data_fit = data[:,(t_fstart<=data[0])&(data[0]<=t_fend)]
        R2fit = mcreep.fit.coefficient_of_determination(self, par, data_fit)
        R2all = mcreep.fit.coefficient_of_determination(self, par, data)
        return(R2fit, R2all)
    
    def recalc_and_save_fitting_results(
            self, datafile, par, R2fit, R2all, t_start):
        '''
        Recalculate and save the results of fitting to Model object.
        The fitting results are saved in two object properties:
        
        * self.table_of_results = the results of fitting
        * self.table_of_results_evp = recalculated results for EVP models
        
        The recalculations convert fitting/regression parameters to
        final compliances of EVP models.
        
        The saved results in Model object can be printed and saved using
        another property mcreep.model.Model.final_report.
        
        Parameters
        ----------
        
        datafile : str
            Datafile containing creep data.
            In this function it is used just for output;
            the (shortened) datafile name denotes the processed dataset.
        
        par : list of floats
            Result of fitting procedure;
            usually the result of procedure mcreep.fit.fit.
        
        R2fit, R2all : float, float
            Coefficients of determination for fitting interval and all data;
            output of the function mcreep.model.Model.calculate_statistics.
        
        t_start : float
            Starting time of fitting.
            In this procedure it is used for calculation of
            RCF = Ramp Correction Factors of EVP models
            (see the notes inside the code).
        
        Returns
        -------
        None
            * Formally it returns self, but this is not necessary.
            * The result are fitting results saved in self = in Model object.

        '''
        # (0) Define function that calculates RCF = ramp correction factors
        # (RCV = rho = constants employed in EVP recalculations below
        # {Mencik 2011} = Polymer Testing 30 (2011) 101–109; Eq.(9) at p.103.
        # (a) tR = time of ramping = until the maximum F is reached
        # (tR is used in rho definition and EVP calculations below
        tR = t_start
        # (b) rho = ramp correction factor, calculated from tau and tR
        def rho(tau): return( (tau/tR) * (math.exp(tR/tau) - 1) )
        # (1) Models
        # (A) Empirical models
        # (Aa) PL = Power Law
        if self.fname == 'power_law':
            results1 = {'C':par[0], 'n':par[1]}
        # (Ab) NL = Nutting's Law
        elif self.fname == 'nutting_law':
            results1 = {'e0':par[0],'C':par[1], 'n':par[2]}
        # (B) EVP models
        elif self.fname.startswith('evp'):
            # Add parameters that are common to all EVP models
            const = self.EPAR.const
            B0,Cv = par[0:2]
            results1 = {'const':const, 'B0':B0, 'Cv':Cv}
            # Prepare parameter rtimes (it will be used many times)
            rtimes = self.rtimes
            # (Ba) EVP: S+D+1KV
            if self.fname == 'evp_s_d_1kv':
                # Get parameters from par and/or rtimes
                # Two specific tricks for EVP1 (unlike EVP2,EVP3)
                # a) we use tau1=rtimes[0] ..to get just one float, not array
                # b) we use D1 = par[-1] ....to get just one float, not array
                # * Note: for EVP2,EVP3 (see below) it works like:
                #   ... D1,D2 = par[-2:] ....list of floats (from array slice)
                #   ... tau1,tau2 = rtimes ..list of floats (from whole array)
                # * Generalization (important to avoid sneaky errors):
                #   To get single float we cannot use whole array/array slice
                #   because this returns array of floats!
                if rtimes: D1 = par[-1]; tau1=rtimes[0]
                else: D1,tau1 = par[-2:]
                # Save FITTED compliances to results1
                results1 = {**results1, 'D1':D1}
                # Calculate and save FINAL compliances to results2
                if self.EPAR.etype == 'Tensile':
                    # calculation for tensile experiments => straightforward
                    C0 = results1['B0'] - D1
                    results2 = {'C0':C0, 'Cv':Cv, 'C1':D1}
                else:
                    # calculation for indentation experiments => uses RCF=rho
                    C1 = results1['D1']/rho(tau1)
                    C0 = results1['B0'] + results1['Cv']*tR/2 - C1
                    results2 = {'C0':C0, 'Cv':Cv, 'C1':C1}
                # Add retardation times to both results1 and results2
                results1 = {**results1, 'tau1':tau1}
                results2 = {**results2, 'tau1':tau1}
            # (Bb) EVP: S+D+2KV
            elif self.fname == 'evp_s_d_2kv':
                # Get parameters from par and/or rtimes
                if rtimes: D1,D2 = par[-2:]; tau1,tau2 = rtimes
                else: D1,D2,tau1,tau2 = par[-4:]
                # Save FITTED compliances to results1
                results1 = {**results1, 'D1':D1, 'D2':D2}
                # Calculate and save FINAL compliances to results2
                if self.EPAR.etype == 'Tensile':
                    # calculation for tensile experiments => straightforward
                    C0 = results1['B0'] - D1 - D2
                    results2 = {'C0':C0, 'Cv':Cv, 'C1':D1, 'C2':D2}
                else:
                    # calculation for indentation experiments => uses RCF=rho
                    C1,C2 = D1/rho(tau1),D2/rho(tau2)
                    C0 = results1['B0'] + results1['Cv']*tR/2 - C1 - C2
                    results2 = {'C0':C0, 'Cv':Cv, 'C1':C1, 'C2':C2}
                # Add retardation times to both results1 and results2
                results1 = {**results1, 'tau1':tau1, 'tau2':tau2}
                results2 = {**results2, 'tau1':tau1, 'tau2':tau2}
            # (Bc) EVP: S+D+3KV
            elif self.fname == 'evp_s_d_3kv':
                # Get parameters from par and/or rtimes
                if rtimes: D1,D2,D3 = par[-3:]; tau1,tau2,tau3 = rtimes
                else: D1,D2,D3,tau1,tau2,tau3 = par[-6:]
                # Save FITTED compliances to results1
                results1 = {**results1, 'D1':D1, 'D2':D2, 'D3':D3}
                # Calculate and save FINAL compliances to results2
                if self.EPAR.etype == 'Tensile':
                    # calculation for tensile experiments => straightforward
                    C0 = results1['B0'] - D1 - D2 - D3
                    results2 = {'C0':C0, 'Cv':Cv, 'C1':D1, 'C2':D2, 'C3':D3}
                else:
                    # calculation for indentation experiments => uses RCF=rho
                    C1,C2,C3 = D1/rho(tau1),D2/rho(tau2),D3/rho(tau3)
                    C0 = results1['B0'] + results1['Cv']*tR/2 - C1 - C2 - C3 
                    results2 = {'C0':C0, 'Cv':Cv, 'C1':C1, 'C2':C2, 'C3':C3}
                # Add retardation times to both results1 and results2
                results1 = {**results1, 'tau1':tau1, 'tau2':tau2, 'tau3':tau3}
                results2 = {**results2, 'tau1':tau1, 'tau2':tau2, 'tau3':tau3}
            # (Bd) Some other/unknown EVP model...
            else: sys.exit('Unknown EVP model!')
        # (D) Completely unknown model...
        else: sys.exit('Unknown model!')
        # (2) Save {results1} to self.table_of_results
        # (a) Add statistics
        results1 = {**results1, 'R2fit':R2fit, 'R2all':R2all}
        # (b) Add/save parameters to  self.table_of_results
        # onvert datafile to pure/string name without path
        # (Reason: datafile will be used as a (string) index of the DataFrame
        datafile = Path(datafile).name
        # (c) Add the new row (created above) to the DataFrame
        # (df.loc[new_index] => new row in df; here [new_index] = [datafile]
        # (other methods of adding rows to df - more difficult, bad indexing
        self.table_of_results.loc[datafile] = results1
        # (3) Save {results2} to self.table_of_results_evp (EVP models only)
        if self.fname.startswith('evp'):
            self.table_of_results_evp.loc[datafile] = results2
        # (4) Return is not necessary (data have been stored in self directly)
        return(self)
    
    def final_report(self, output_file=None):
        '''
        Print and save final report = summarize the results of fitting.
        The report is calls the following methods
        of Model object:
        
        * Model.describe = basic text description of the model.
        * Model.print_table_of_results = print results of fitting.
        * Model.print_table_of_results_evp = print addidional results
          of fitting for EVP models.
          
        The results are both printed and saved in the text file.

        Parameters
        ----------
        output_file : str, optional
            Name of the output text file, into which the results are saved.
            If the parameter is given

        Returns
        -------
        None
            Output is the printed report
            and saved file with the results of fitting.

        '''
        # (0) Get name of output file and open it
        # ...if output_file was not given, create default = input_file.txt
        if output_file == None: output_file = sys.argv[0] + '.txt'
        fh = open(output_file, 'w')
        # (1) Save basic description of the model to file
        self.describe(fh)
        # (2) Print + save basic table of results
        # (for all models, this table contains fitting + statistics
        self.print_table_of_results()
        self.save_table_of_results(fh)
        # (3) Print + save additional table of results for EVP models
        # (only for EVP: recalculation of parameters to predict J(t) = f(t)
        if self.fname.startswith('evp'):
            self.print_table_of_results_evp()
            self.save_table_of_results_evp(fh)
        # (4) End of report, close output file
        fh.close()
        
    def print_table_of_results(self):
        '''
        Print table of results as text
        (this is done by a small trick employing pandas.DataFrame).
        '''
        print('\nFitting results and statistics:\n')
        s = self.table_of_results.to_string(float_format = '%.4f')
        print(s)
        
    def print_table_of_results_evp(self):
        '''
        Print table of additional result of EVP models as text
        (this is done by a small trick employing pandas.DataFrame).
        '''
        print('\nFinal compliances and retardation times of EVP model:\n')
        s = self.table_of_results_evp.to_string(float_format = '%.4f')
        print(s)
        
    def save_table_of_results(self, fh):
        '''
        Save table of results as text
        (this is done by a small trick employing pandas.DataFrame).
        '''
        s = 'Fitting results & statistics:\n\n'
        s = s + self.table_of_results.to_string(float_format = '%.6f')
        fh.write(s)
        
    def save_table_of_results_evp(self, fh):
        '''
        Save table of additional results of EVP models as text
        (this is done by a small trick employing pandas.DataFrame).
        '''
        s = '\n\nFinal compliances and retardation times of EVP model:\n\n'
        s = s + self.table_of_results_evp.to_string(float_format = '%.6f')
        fh.write(s)



# Python package initialization file.

'''
MCREEP package
-------------
Fitting of various pre-defined models to creep data.

* The creep data:
    - a text file with two columns
    - the 1st column = time, the 2nd column = deformation
* The creep data may come from various experiments:
    - tensile creep => deformation(t) = strain(t) = epsilon(t)
    - indentation creep => deformation(t) = penetration_depth(t) = h(t)
* The pre-defined models within the pagkage are:
    - simple empirical models = Power law, Nutting's law
    - phenomenological models = elasto-visco-plastic models = EVP models
* Technical notes:
    - tensile creep is usually measured in macroscale
    - indentation creep is usually measured in micro/nanoscale
    - the models used in this package work for all three scales

Typical (simple/minimalistic but complete/working) example is as follows:
    
>>> # (0) Import modules
>>> import mcreep.const, mcreep.func, mcreep.model
>>>
>>> # (1) Define experiment, data and plot parameters
>>> EPAR = mcreep.const.Experiment(etype='Vickers', F=100)
>>> DPAR = mcreep.const.DataParameters(deformation_to_um=0.001)
>>> PPAR = mcreep.const.PlotParameters(xlabel='t [s]', ylabel='h [um]')
>>>
>>> # (2) Define model (using functions prepared in mcreep.func)
>>> MODEL = mcreep.model.Model(EPAR, DPAR, PPAR,
>>>     name='S+D+2KV', func=mcreep.func.evp_s_d_2kv, rtimes=[5,20])
>>>
>>> # (3) Run the whole calculation
>>> MODEL.run('pe1.txt', t_start=2.2, t_hold=99)
>>>
>>> # (4) Create report = save and print fitting results
>>> MODEL.final_report()
'''

__version__ = "1.1.6"



'''
mcreep.const
------------
Global constants for package mcreep.

* Global constants are saved in the form of objects.
* Reason: Object parameters can be [re]defined during initialization.
'''

from math import pi,sqrt,tan
import matplotlib.pyplot as plt

class Experiment:
    '''
    Experiment = object describing the creep experiment.
    
    Parameters
    ----------
    etype : str; which takes one o the following values
        'Tensile'   = (macroscale) tensile creep experiment.
        'Vickers'   = indentation creep with Vickers tip.
        'Berkovich' = indentation creep with Berkovich tip.
        'Spherical' = indentation creep with spherical tip.
    F : float; optional, default is None
        Loading force during indentatin experiments in [mN].
        Relevant (and obligatory) for etype='Vickers/Berkovich/Spherical'.
    R : float; optional, the default is None
        Radius of spherical tip in [um].
        Relevant (and obligatory) for etype='Spherical'.
    sigma : float; optional, default is None
        Stress during tensile experiments in [GPa].
        Relevant (and obligatory) for etype='Tensile'.
    
    Additional parameters
    ---------------------
    * m,K = additional constants that are used
      for correct recalculations of indentation creep data.
      These parameters are calculated automatically,
      based on the value of parameter `etype`.
    * const = yet additional constant that is used
      for correct fitting of EVP models
      to both indentation and tensile creep data with EVP models.
        - EVP in general:
          `def(t) = const * { A0 + Av*t + Sum[f(Ai,ti)] }`
        - EVP in tensile creep:
          `def(t) = epsilon(t)` and `const = sigma[GPa]`     
        - EVP in indentation creep:
          `def(t) = [h(t)**m]/K` and `const = F[mN]`
        - Note for indentation creep:
            - Mencik_2009: `def(t) = h(t)**m` and `const = F*K`
            - Here (see above): `def(t) = [h(t)**m]/K` and `const = F[mN]`
            - Reason for the difference:
                - more consistent treatment of PL x NL x EVP models
                - for all models we fit: `[h(t)**m]/K`
    
    Sample calculation of sigma
    ---------------------------
    * `sigma[GPa] = (load[kg]*SA*g[m/s2]) / (W[mm]*T[mm]/1e6) / 1e9`
    * `SA` = stress amplifier (optional lever amplifying the load)
    * `g` = net acceleration = gravity of Earth = 9.81[m/s2]
    * `W,T` = width and thickness of testing specimen in [mm]
    * `1e6` = recalculate W*T[mm2] -> [m2]
    * `1e9` = recalculate final result [Pa] -> [GPa]
    
    '''
    def __init__(self, etype, F=None, R=None, sigma=None):
        # Docstring for __init__ is given above in the class definition.
        # Reason: consistent help in Spyder and Pdoc.
        # -----        
        
        # (1) Initialize basic parameters
        self.etype = etype
        self.F = F
        self.R = R
        self.sigma = sigma
        
        # (2) Calculate additional parameters: K,m
        # (parameters K,m calculated according to {Mencik 2011}
        # {Mencik 2011} = Polymer Testing 30 (2011) 101–109; Eq.(9) at p.103.
        if self.etype == 'Tensile':
            self.K = 1
            self.m = 1
        elif self.etype in ['Vickers','Berkovich']:
            alpha = 70.3 * pi/180
            self.K = pi/(2*tan(alpha))
            self.m = 2
        elif self.etype == 'Spherical':
            self.K = 3/(4*sqrt(self.R))
            self.m = 3/2
        
        # (3) Calculate additional parameter: const
        # (const is employed in EVP models,
        # (...in which it is the 1st multiplicative constant
        # (...that differs for tensile and indentation experiments;
        # (this is my generalization, which enables that the package
        # (...can be used for both tensile and indentation experiments        
        if self.etype == 'Tensile':
            self.const = self.sigma
        elif self.etype in ['Vickers','Berkovich','Spherical']:
            # Mencik_2009:      `def(t) = h(t)**m` and `const = F*K`
            # Here (see above): `def(t) = [h(t)**m]/K` and `const = F[mN]`
            # Reason for the difference:
            # consistent treatment of PL x NL x EVP models.
            self.const = self.F


class DataParameters:
    '''
    DataParameters = object defining format of a creep datafile.
    
    Assumptions:
    * creep datafile is a TXT file containing data in columns;
    * one of the columns contains time, some other contains deformation.
    
    Parameters
    ----------
    usecols : list with two integer values; optional, the default is [0,1]
        This parameter is passed to numpy.loadtxt.
        * The 1st value of the list = column with times.
        * The 2nd value of the list = column with deformations.
    
    comments : string or sequence of strings; optional, the default is '#'
        This parameter is passed to numpy.loadtxt.
        The character(s) indicate comment/ignored lines in input datafile.
    
    skiprows : integer; optional, the default is 1
        This parameter is passed to numpy.loadtxt.
        The first {skiprows} lines are skipped.
    
    time_to_seconds : float; optional, the default is 1
        Multiplicative constant, which converts time values to seconds;
        this conversion is necessary for the following calculations.
    
    deformation_to_um : float; optional, the default is 1
        Multiplicative constant, which converts deformation to micrometers;
        this conversion is necessary for the following calculations. 
    '''
         
    def __init__(self, usecols=[0,1], comments='#', skiprows=0,
                 time_to_seconds=1, deformation_to_um=1):
        # Docstring for __init__ is given above in the class definition.
        # Reason: consistent help in Spyder and Pdoc.
        # -----        
        
        self.usecols = usecols
        self.comments = comments
        self.skiprows = skiprows
        self.time_to_seconds = time_to_seconds
        self.deformation_to_um = deformation_to_um
        
class PlotParameters:
    '''
    PlotParameters = object defining local+global parameters for plotting.
    
    Parameters
    ----------
    
    xlabel, ylabel : str, str
        Labels for X and Y axis.
    
    logscale : bool; optional, the default is False
        If logscale==True, both X and Y axes are in logarithmic scale.
    
    e_to_percent : bool; optional, the default is True
        Relevant only to tensile experiments.
        If true, the values of elongation are multiplied by 100,
        i.e. they are converted from epsilon[] to epsilon[%].
        
    rcParams : dict; optional, the default is empty dictionary {}
        The dictionary shoud be formatted for mathplotlib.pyplot.rcParams.
        The argmument is passed to matplotlib.pyplot.
        The initialization procedure creates some default rcParams.
        This argument can override this pre-defined parameters,
        i.e. the default is created anyway
        and then (possibly) supplemented by rcParams argument.
    
    showfigs : bool; optional, the default is True.
        If showfigs==True, the figures are shown + saved in files,
        which is default behavior, suitable for running the script in Spyder.
        If showfigs==False, the figures are just saved in files,
        which is an option, suitable for running the script from CLI.
        
    ax : matplotlib Axes object, the default is None
        * If ax == None, create and save results as a single plot,
          which is a typical usage.
        * If ax is defined, create the plot within given ax object,
          which can combined with `fig,ax = plt.subplots()`
          in order to create multile figures.
          
    legend_coordinates : None or tuple of two floats, optional, default is None
        If not given (typical case)
        some suitable default position of legend argument will be set.
        If the argument is given,
        the legend is placed at given position within the graph;
        *legend_coordinates* = upper left corner, in fractional coordinates.
    '''
    
    def __init__(self, xlabel, ylabel,
                 logscale=False, e_to_percent=True, rcParams={},
                 showfigs=True, ax=None, legend_coordinates=None):
        # Docstring for __init__ is given above in the class definition.
        # Reason: consistent help in Spyder and Pdoc.
        # -----        
        
        # (1) Initialize basic parameters
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.logscale = logscale
        self.e_to_percent = e_to_percent
        self.rcParams = rcParams
        self.showfigs = showfigs
        self.ax = ax
        self.legend_coordinates = legend_coordinates
        
        # (2) Set global plot settings using rcParams
        PlotParameters.set_default_rcParams(rcParams)
        
    @classmethod
    def set_default_rcParams(cls, my_rcParams={}):
        '''
        A class method defining global plot parameters (plt.rcParams).
        
        Parameters
        ----------
        my_rcParams : dict
            The dictionary re-defines selected plt.rcParams keys.
            
            Example:
                
            >>> PlotParameters.set_default_rcParams({'figure.dpi':500})
            
        Returns
        -------
        None
            The function does not return anything,
            BUT it re-defines the global variable rcParams.

        Notes
        -----
        * This is a @classmethod (because it is used within the whole class)
          but it could be a @staticmethod as well (because it does not use
          cls variable in fact).
        * The method is employed in two ways:
            - Standard usage of MCREEP package: default rcParams are used
              (and possibly modified) in objects of PlotParameters class.
            - Special usage of MCREEP (more figures, multiplots): default
              rcParams are used when definining the axes of (multiple) figures.
        '''
        
        # (1) Set default rcParams
        # (Hardcoded, suitable default for standard plots
        plt.rcParams.update({
            'figure.figsize'     : (8/2.54,6/2.54),
            'figure.dpi'         : 500,
            'font.size'          : 7,
            'lines.linewidth'    : 0.8,
            'axes.linewidth'     : 0.6,
            'xtick.major.width'  : 0.6,
            'ytick.major.width'  : 0.6,
            'grid.linewidth'     : 0.6,
            'grid.linestyle'     : ':'})
        
        # (2) Update default with argument rcParams, if it was given
        # (User-defined in the main program, if necessary
        # (Useful namely for multiplots
        plt.rcParams.update(my_rcParams)



'''
mcreep.io
-----------
Input/output functions for package mcreep.    
'''

import sys
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from pathlib import Path

def read_datafile(MODEL, datafile, t_start, t_hold):
    '''
    Read datafile containing creep data.
    
    Parameters
    ----------
    MODEL : MODEL object
        The object keeps experiment parameters in MODEL.EPAR property.
        We use the object to convert time and deformation to correct units.
    datafile : string or pathlib object
        Name of datafile = a text file containing
        at least two columns (time,deformation);
        see section {Expected format of the datafile}
        below for more information.
    t_start : float
        Start of step II = holding step = period at which F = Fmax.
    t_end : float
        End of step II = holding step = period at which F = Fmax.
        
    Returns
    -------
    2D numpy array.
    
    Expected format of the datafile
    -------------------------------
    * Two colums (time,deformation) separated by whitespace.
        - 1st column = t[s] = time in seconds
        - 2nd column = deformation[um] = deformation in micrometers
            - in tensile experiments: deformation = strain
            - in indentation experiments: deforamtion = penetration depth
    * The datafile is read by means of numpy.loadtxt function
      and some parameters can be adjusted by means of MODEL argument
      (which contains DataParameters object describing the input data):
        - see source code of this function
        - and mcreep.const.DataParameters object
        - for example, it is possible to specify column numbers
          if the datafile contains more columns than required
    '''
    # Read file to numpy array and try to catch possible errors/exceptions
    try:
        # read h-t file
        data = np.loadtxt(
            datafile, unpack=True,
            usecols = MODEL.DPAR.usecols,
            skiprows = MODEL.DPAR.skiprows,
            comments = MODEL.DPAR.comments)
        # take just section II of h-t curve
        # section I = loading  = (0..t_start] ...without 0 due to logarithms
        # section II = holding = [t_start..t_hold] ...time of maximal loading
        data = data[:,(data[0]>=t_start) & (data[0]<=(t_start+t_hold))]
        # recalculate time and deformation to [s] and [um], respectively
        data[0] = data[0] * MODEL.DPAR.time_to_seconds
        data[1] = data[1] * MODEL.DPAR.deformation_to_um
    except OSError as err:
        print('OSError:', err)
        sys.exit()
    except ValueError:
        print('ValueError: probably a wrong format of the file.')
        print('Expected: TXT file with two columns t[s], def[length_units]')
        print('...where def[length_units] is conveted to def[um]')
        print('...with a user-defined constant MODEL.DPAR.deformation_to_um')
        sys.exit()
    # Return final 2D numpy array
    # (1st column = t[s], 2nd col = def[um], range = [t_start;t_start+t_hold]
    return(data)
    
def read_list_of_datafiles(filename):
    '''
    Read a text file containing the list of datafiles with loading times.

    Parameters
    ----------
    filename : str
        Name of the text file, which contains header and two columns.
        The header = the 1s of the file = column names: Datafile t_start
        The 1st column = datafile names (relative paths to current dir).
        The 2nd column = real loading times for the individual datafiles.

    Returns
    -------
    datafiles : list of strings
        List of datafiles names (relative paths).
    loading_times : list of floats
        List of real loading times for each datafile.
    '''
    # Read filename = text file with dafile names and loading times
    # (Format of the file is specified above in function docstring
    # (It is a simple text file, which can be prepared manually
    # (Typically, the text file is made with an auxiliary script and edited
    # (See examples on www for more information
    df = pd.read_csv(filename, sep='\s+')
    # If the file exists and has the required format,
    # we simply convert the two columns to lists and return them
    datafiles = list(df.Datafile)
    loading_times = list(df.t_start)
    # Return the list of datafiles and loading times
    return(datafiles, loading_times)

def print_fitting_result(MODEL, datafile, par, cov):
    '''
    Print ONE result of fitting
    (i.e. print fitting results for one/currently processed datafile).

    We note that the complete results of each fitting (for each datafile)
    are kept in mcreep.model.Model object and usually reported/printed
    together at the very end of the whole processing.
    
    Exception: Covariance matrixes of each fitting are just printed here,
    NOT kept in mcreep.model.Model object and NOT saved to file.
    Moreover, they are printed just on request (MODEL.print_covariances=True).
    They are needed just occasionally, and if really needed, the values
    can be copy+pasted from stdout and saved to file.

    Parameters
    ----------
    MODEL : mcreep.model.Model object
        This object contains all properties needed to print the result.
        
    datafile : str
        Name of the datafile;
        in this procedure, it is just printed to stdout
        in order to denote the source of the fitting results.
        
    par : list of floats
        List of optimized parameters of given fitting function;
        output from the procedure mcreep.fit.fit.
        
    cov : 2D-array of floats
        Covariance matrix for all fitted/regression parameters;
        supplementary output from the procedure mcreep.fit.fit.

    Returns
    -------
    None
        The output are results printed on stdout.
    '''
    # (1) Convert parameters array to tuple => suitable for formatted printing 
    par = tuple(par)
    # (2) Print datafile name (this is the same for all models)
    print(f'{datafile} ', end='')
    # (3) Print fitting/regression parameters...
    # * The parameters are printed in the format/order specific to given model.
    if MODEL.fname == 'power_law':
        print('[C,n]: %8.4f %8.4f' % par)
    elif MODEL.fname == 'nutting_law':
        print('[e0,C,n]: %8.4f %8.4f %8.4f' % par)
    elif MODEL.fname == 'evp_s_d_1kv':
        if MODEL.rtimes:
            print('[B0,Cv,D1]: %8.4f %8.4f %8.4f' % par)
        else:
            print('[B0,Cv,D1,tau1]: %8.4f %8.4f %8.4f %6.2f' % par) 
    elif MODEL.fname == 'evp_s_d_2kv':
        if MODEL.rtimes:
            print('[B0,Cv,D1,D2]: ', end='')
            print('%8.4f %8.4f %8.4f %8.4f' % par)
        else:
            print('[B0,Cv,D1,D2,tau1,tau2]: ', end='')
            print('%8.4f %8.4f %8.4f %8.4f %6.2f %6.2f' % par) 
    elif MODEL.fname == 'evp_s_d_3kv':
        if MODEL.rtimes:
            print('[B0,Cv,D1,D2,D3]: ',end='')
            print('%8.4f %8.4f %8.4f %8.4f %8.4f' % par)
        else:
            print('[B0,Cv,D1,D2,D3,tau1,tau2,tau3]: ', end='')
            print('%8.4f %8.4f %8.4f %8.4f %8.4f %6.2f %6.2f %6.2f' % par)
    else:
        print('Warning: unknown model during printing!')
        print(par)
    # (4) Print covariance matrix showing independence of parameters...
    # * The cov.matrix is printed on request (MODEL.print_covariances=True).
    # * The order of rows/columns of the cov.matrix the same like in item (3).
    #   => therefore, we do not re-print the parameter names like in item (3).
    # * Fitting/regression parameters are usually saved to file
    #   (by means of MODEL.final_report() method).
    # * Covariance matrixes is just printed to stdout here.
    #   (they can be copy+pasted from stdout and saved to file manually).
    if MODEL.print_covariances == True:
        print('\nCovariance matrix of all parameters after fitting:')
        for row in cov:
            for val in row:
                print(f'{val:10.6f}', end='')
            print()

def plot_fitting_result(MODEL, datafile, data, par):
    '''
    Plot the result of fitting (single plot OR axes defined within MODEL).
        
    Parameters
    ----------
    MODEL : mcreep.model.Model object
        This object contains all properties needed to print the result.
        
    datafile : str or path-like object
        Name of the datafile containing input data.
        * Important: We use this argument just for creating
          the name of the output graph => {datafile}.png.
        * If MODEL.PPAR.ax == None: the output goes to single PNG file,
          whose name is created as `datafile.PNG`.
        * If MODEL.PPAR.ax != None: the output goes to pre-prepared axes
          object, and datafile parameter is ignored.
    
    data : 2D numpy array
        Creep data read from the currently processed datafile.
        It is expectted 
        
    par : list of floats
        List of optimized parameters of given fitting function;
        output from the procedure mcreep.fit.fit.

    Returns
    -------
    None
        The output is either single plot OR axes in a multiplot.
        * If MODEL.PPAR.ax == None:
          a single output graph is shown (option) + saved (always).
          The name of the output graph is {datafile}.png.
        * If MODEL.PPAR.ax == axes_object
          a graphs is created within the pre-defined axes_object.
          This can be employed in creating user-defined multiplots.
    '''
    # (1) Read data
    X,Y = data
    # (2) Calculate fitted data
    # Trick: *(list(par)) = convert saved parameters to list and expand
    Yfit = MODEL.func(X, *(list(par)))
    # (3) For tensile experiments, convert elongation to % if required.
    if MODEL.EPAR.etype == 'Tensile' and MODEL.PPAR.e_to_percent == True:
        Y = Y*100
        Yfit = Yfit*100
    # (4) Recalculate fitting function
    # Reason: Tensile x Indentation experiments fit different deformations...
    # More details => see the description of the recalculating fucntion.
    Yfit = recalculate_fitted_data(MODEL.EPAR, Yfit)
    # (5) If required, convert data to logscale
    # AND change x/ylabels accordingly.
    if MODEL.PPAR.logscale == True:
        # (a) Convert data to logscale
        X = np.log10(X)
        Y = Y = np.log10(Y)
        Yfit = np.log10(Yfit)
        # (b) change x/ylabels
        # Trick: this function is called repeatedly
        # => check if the change of x/ylabes has not been already done!
        if not(MODEL.PPAR.xlabel.startswith('log')):
            MODEL.PPAR.xlabel = 'log('+MODEL.PPAR.xlabel+')'
        if not(MODEL.PPAR.ylabel.startswith('log')):
            MODEL.PPAR.ylabel = 'log('+MODEL.PPAR.ylabel+')'
    # (6) Define suitable default for legend coordinates
    # (in a typical case when the legend coordinates were NOT given as argument
    if MODEL.PPAR.legend_coordinates == None:
        my_legend_coordinates = (0.52,0.28)
    # (7) Create plot
    # (user-adjustable + global plot settings are saved in PPAR object
    # (MODEL.PPAR object is an instance of mcreep.const.PlotParameters class
    # (7a) No axes_object was given as argument => create+save single plot
    if MODEL.PPAR.ax == None:
        # (a) Create name of output graph => name of the plot to save.
        # (output graph = MODEL.output_dir/datafile.png
        output_filename = Path(datafile).name + '.png'
        output_graph = Path(MODEL.output_dir, output_filename)
        # (b) Create the plot
        plt.plot(X, Y, color='orange', label='Experiment')
        plt.plot(X, Yfit, 'k:', label=MODEL.name)
        plt.xlabel(MODEL.PPAR.xlabel)
        plt.ylabel(MODEL.PPAR.ylabel)
        plt.grid()
        plt.legend(loc='upper left', bbox_to_anchor=my_legend_coordinates)
        plt.tight_layout()
        # (c) Save the plot
        plt.savefig(output_graph)
        # (d) Show and close the plot
        # (Default is showfigs=True => figs shown+closed - suitable for Spyder
        # (Option is to set showfigs==False => just close - suitable for CLI
        # (Technical notes:
        # ( * plt.show ...show (and close) plot in Spyder
        # ( * plt.close ..close plot explicitly
        # (   needed for multiple plots + showfigs==False
        # (   otherwise all plots would be drawn into just one figure
        if MODEL.PPAR.showfigs == True:
            plt.show()
        plt.close()
    # (7b) axes object was given as argument => create plot within the axes
    # (this is employed when creating multiple plots or multiplots
    else:
        ax = MODEL.PPAR.ax
        ax.plot(X, Y, color='orange', label='Experiment')
        ax.plot(X, Yfit, 'k:', label=MODEL.name)
        ax.set_xlabel(MODEL.PPAR.xlabel)
        ax.set_ylabel(MODEL.PPAR.ylabel)
        ax.grid()
        ax.legend(loc='upper left', bbox_to_anchor=my_legend_coordinates)
        
def recalculate_fitted_data(EPAR, Y_orig):
    '''
    Recalculation of the fitted/calculated data before plotting.
   
    Parameters
    ----------
    EPAR : Experiment object
        Object with experimental parameters.
        In this function, we need (EPAR.m,EPAR.K) for Y-data recalculation.
    Y_orig : 1D numpy array.
        Original Y-data, to which f(t) was fitted.
        The data have to be recalculated, because...
        
        * The models could be fitted to tensile or indentation creep data.
            * Tensile creep data were not modified: deformation(t) = epsilon(t)
            * Indentation creep data were modified: deformation(t) = (h**m)/K)
        * The modification is necessary in order to achieve
          compatibility between tensile and indentation creep results.
            * More precisely, indentation creep data are recalculated
              so that the results
              from (specific) fitting of models to indentation creep data
              are compatible with the results
              from (standard) fitting of models to tensile creep data.
        * For INDENTATION CREEP, the modification means that...
            * Model was not fitted to [h(t)] but to [h(t)**m]/K
            * Fitted data calculated from model are [h(t)**m]/K
            * These data have to be calculated to [h(t)] before plotting:
                * `Y_orig = [h(t)**m]/K`
                * `h(t) = (Y_orig * K) ** (1/m)`

    Returns
    -------
    Y_recalc : 1D numpy array
        Recalculated Y-data for plotting.
        The recalculated data are:
            * epsilon(t) for tensile creep
            * h(t) for indentation creep
    
    '''
    if EPAR.etype == 'Tensile':
        Y_recalc = Y_orig
    else:
        Y_recalc = (Y_orig * EPAR.K) ** (1/EPAR.m)
    return(Y_recalc)


import setuptools

def get_version():
    with(open("src/mcreep/__init__.py", "r")) as fh:
        for line in fh:
            if line.startswith("__version__"):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
        else:
            raise RuntimeError("Unable to find version string.")
                
def get_long_description():
    with open("README.md", "r") as fh: description = fh.read()
    return(description)
    
setuptools.setup(
    name="mcreep",
    version=get_version(),
    author="Mirek Slouf",
    author_email="mirek.slouf@gmail.com",
    description="Fitting of various pre-defined models to creep data.",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/mirekslouf/mcreep/",
    project_urls={
        "Documentation": "https://mirekslouf.github.io/mcreep/"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"],
    license='MIT',
    package_dir={"":"src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    include_package_data=True)
